from datetime import date, datetime
import unittest

from pandas import Timedelta
from owlapy.model.providers import OWLDatatypeMaxInclusiveRestriction, OWLDatatypeMinInclusiveRestriction, \
    OWLDatatypeMinMaxExclusiveRestriction, OWLDatatypeMinMaxInclusiveRestriction

import owlready2
import owlapy.owlready2.utils
from owlapy.model import OWLObjectProperty, OWLNamedIndividual, OWLThing, OWLClass, OWLObjectUnionOf, \
    OWLObjectIntersectionOf, OWLObjectSomeValuesFrom, OWLObjectComplementOf, IRI, OWLDataAllValuesFrom, \
    OWLDataComplementOf, OWLDataHasValue, OWLDataIntersectionOf, OWLDataProperty, OWLDataSomeValuesFrom, \
    OWLDataUnionOf, OWLLiteral, BooleanOWLDatatype, DoubleOWLDatatype, IntegerOWLDatatype, OWLDataOneOf, \
    OWLDataExactCardinality, OWLDataMaxCardinality, OWLDataMinCardinality, OWLObjectExactCardinality, \
    OWLObjectMaxCardinality, OWLObjectMinCardinality, OWLObjectHasValue, OWLObjectAllValuesFrom, \
    OWLObjectOneOf, DateOWLDatatype, DateTimeOWLDatatype, DurationOWLDatatype

from owlapy.owlready2 import OWLOntologyManager_Owlready2, OWLReasoner_Owlready2
from owlapy.owlready2.temp_classes import OWLReasoner_Owlready2_TempClasses


class Owlapy_Owlready2_Test(unittest.TestCase):
    def test_sub_classes(self):
        NS = "http://example.com/father#"
        mgr = OWLOntologyManager_Owlready2()
        onto = mgr.load_ontology(IRI.create("file://KGs/father.owl"))
        reasoner = OWLReasoner_Owlready2(onto)

        classes = frozenset(reasoner.sub_classes(OWLClass(IRI.create(NS, "person"))))
        target_classes = frozenset((OWLClass(IRI.create(NS, "male")), OWLClass(IRI.create(NS, "female"))))
        self.assertEqual(classes, target_classes)

    def test_sub_object_properties(self):
        NS = "http://www.biopax.org/examples/glycolysis#"
        mgr = OWLOntologyManager_Owlready2()
        onto = mgr.load_ontology(IRI.create("file://KGs/Biopax/biopax.owl"))
        reasoner = OWLReasoner_Owlready2(onto)

        participants = OWLObjectProperty(IRI.create(NS, 'PARTICIPANTS'))
        target_props = frozenset({OWLObjectProperty(IRI(NS, 'COFACTOR')),
                                  OWLObjectProperty(IRI(NS, 'CONTROLLED')),
                                  OWLObjectProperty(IRI(NS, 'CONTROLLER')),
                                  OWLObjectProperty(IRI(NS, 'LEFT')),
                                  OWLObjectProperty(IRI(NS, 'RIGHT'))})
        self.assertEqual(frozenset(reasoner.sub_object_properties(participants, direct=True)), target_props)
        self.assertEqual(frozenset(reasoner.sub_object_properties(participants, direct=False)), target_props)

    def test_instances(self):
        NS = "http://example.com/father#"
        mgr = OWLOntologyManager_Owlready2()
        onto = mgr.load_ontology(IRI.create("file://KGs/father.owl"))
        reasoner = OWLReasoner_Owlready2(onto)

        inst = frozenset(reasoner.instances(OWLThing))
        target_inst = frozenset({OWLNamedIndividual(IRI(NS, 'anna')),
                                 OWLNamedIndividual(IRI(NS, 'heinz')),
                                 OWLNamedIndividual(IRI(NS, 'markus')),
                                 OWLNamedIndividual(IRI(NS, 'martin')),
                                 OWLNamedIndividual(IRI(NS, 'michelle')),
                                 OWLNamedIndividual(IRI(NS, 'stefan'))})
        self.assertEqual(inst, target_inst)

        inst = frozenset(reasoner.instances(OWLClass(IRI.create(NS, "male"))))
        target_inst = frozenset({OWLNamedIndividual(IRI(NS, 'heinz')), OWLNamedIndividual(IRI(NS, 'martin')),
                                 OWLNamedIndividual(IRI(NS, 'markus')), OWLNamedIndividual(IRI(NS, 'stefan'))})
        self.assertEqual(inst, target_inst)

    def test_types(self):
        NS = "http://example.com/father#"
        mgr = OWLOntologyManager_Owlready2()
        onto = mgr.load_ontology(IRI.create("file://KGs/father.owl"))
        reasoner = OWLReasoner_Owlready2(onto)

        types = frozenset(reasoner.types(OWLNamedIndividual(IRI.create(NS, 'stefan'))))
        target_types = frozenset({OWLThing, OWLClass(IRI(NS, 'male')), OWLClass(IRI(NS, 'person'))})
        self.assertEqual(target_types, types)

    def test_object_values(self):
        NS = "http://example.com/father#"
        mgr = OWLOntologyManager_Owlready2()
        onto = mgr.load_ontology(IRI.create("file://KGs/father.owl"))
        reasoner = OWLReasoner_Owlready2(onto)

        stefan = OWLNamedIndividual(IRI.create(NS, 'stefan'))
        has_child = OWLObjectProperty(IRI.create(NS, 'hasChild'))

        kids = frozenset(reasoner.object_property_values(stefan, has_child))
        target_kids = frozenset({OWLNamedIndividual(IRI(NS, 'markus'))})
        self.assertEqual(target_kids, kids)

        heinz = OWLNamedIndividual(IRI(NS, 'heinz'))
        no_kids = frozenset(reasoner.object_property_values(heinz, has_child))
        self.assertEqual(frozenset(), no_kids)

    def test_mapping(self):
        NS = "http://example.com/father#"
        mgr = OWLOntologyManager_Owlready2()
        onto = mgr.load_ontology(IRI.create("file://KGs/father.owl"))

        male = OWLClass(IRI.create(NS, 'male'))
        female = OWLClass(IRI.create(NS, 'female'))
        has_child = OWLObjectProperty(IRI(NS, 'hasChild'))

        to_owlready = owlapy.owlready2.utils.ToOwlready2(world=onto._world)

        ce = male
        owlready_ce = onto._onto.male
        self.assertEqual(owlready_ce, to_owlready.map_concept(ce))

        ce = OWLObjectUnionOf((male, female))
        owlready_ce = onto._onto.male | onto._onto.female
        self.assertEqual(owlready_ce, to_owlready.map_concept(ce))

        ce = OWLObjectIntersectionOf((male, female))
        owlready_ce = onto._onto.male & onto._onto.female
        self.assertEqual(owlready_ce, to_owlready.map_concept(ce))

        ce = OWLObjectSomeValuesFrom(has_child, OWLThing)
        owlready_ce = onto._onto.hasChild.some(owlready2.owl.Thing)
        self.assertEqual(owlready_ce, to_owlready.map_concept(ce))

        ce = OWLObjectAllValuesFrom(has_child, male)
        owlready_ce = onto._onto.hasChild.only(onto._onto.male)
        self.assertEqual(owlready_ce, to_owlready.map_concept(ce))

        ce = male.get_object_complement_of()
        owlready_ce = owlready2.Not(onto._onto.male)
        self.assertEqual(owlready_ce, to_owlready.map_concept(ce))

        ce = OWLObjectOneOf([OWLNamedIndividual(IRI.create(NS, 'martin')),
                             OWLNamedIndividual(IRI.create(NS, 'michelle'))])
        owlready_ce = owlready2.OneOf([onto._onto.martin, onto._onto.michelle])
        self.assertEqual(owlready_ce, to_owlready.map_concept(ce))

        ce = OWLObjectMinCardinality(2, has_child, male)
        owlready_ce = onto._onto.hasChild.min(2, onto._onto.male)
        self.assertEqual(owlready_ce, to_owlready.map_concept(ce))

        ce = OWLObjectMaxCardinality(5, has_child, female)
        owlready_ce = onto._onto.hasChild.max(5, onto._onto.female)
        self.assertEqual(owlready_ce, to_owlready.map_concept(ce))

        ce = OWLObjectExactCardinality(3, has_child, OWLThing)
        owlready_ce = onto._onto.hasChild.exactly(3, owlready2.owl.Thing)
        self.assertEqual(owlready_ce, to_owlready.map_concept(ce))

        ce = OWLObjectHasValue(has_child, OWLNamedIndividual(IRI.create(NS, 'markus')))
        owlready_ce = onto._onto.hasChild.value(onto._onto.markus)
        self.assertEqual(owlready_ce, to_owlready.map_concept(ce))

    def test_mapping_data_properties(self):
        NS = "http://dl-learner.org/mutagenesis#"
        mgr = OWLOntologyManager_Owlready2()
        onto = mgr.load_ontology(IRI.create("file://KGs/Mutagenesis/mutagenesis.owl"))

        act = OWLDataProperty(IRI(NS, 'act'))
        charge = OWLDataProperty(IRI(NS, 'charge'))

        to_owlready = owlapy.owlready2.utils.ToOwlready2(world=onto._world)

        # owlready2 defines no equal or hash method for ConstrainedDatatype, just using the __dict__ attribute
        # should be sufficient for the purpose of these tests
        def constraint_datatype_eq(self, other):
            return isinstance(other, owlready2.ConstrainedDatatype) and self.__dict__ == other.__dict__
        setattr(owlready2.ConstrainedDatatype, '__eq__', constraint_datatype_eq)
        setattr(owlready2.ConstrainedDatatype, '__hash__', lambda self: hash(frozenset(self.__dict__.items())))

        ce = OWLDataSomeValuesFrom(act, DateOWLDatatype)
        owlready_ce = onto._onto.act.some(date)
        self.assertEqual(owlready_ce, to_owlready.map_concept(ce))

        res = OWLDatatypeMinInclusiveRestriction(20)
        ce = OWLDataAllValuesFrom(charge, OWLDataComplementOf(res))
        owlready_ce = onto._onto.charge.only(owlready2.Not(owlready2.ConstrainedDatatype(int, min_inclusive=20)))
        self.assertEqual(owlready_ce, to_owlready.map_concept(ce))

        res_both = OWLDatatypeMinMaxExclusiveRestriction(0.5, 1)
        ce = OWLDataAllValuesFrom(charge, OWLDataUnionOf([res, res_both]))
        owlready_ce = onto._onto.charge.only(
            owlready2.Or([owlready2.ConstrainedDatatype(int, min_inclusive=20),
                          owlready2.ConstrainedDatatype(float, min_exclusive=0.5, max_exclusive=1.0)]))
        self.assertEqual(owlready_ce, to_owlready.map_concept(ce))

        res = OWLDatatypeMaxInclusiveRestriction(1.2)
        oneof = OWLDataOneOf([OWLLiteral(2.3), OWLLiteral(5.9), OWLLiteral(7.2)])
        ce = OWLDataAllValuesFrom(charge, OWLDataIntersectionOf([res, oneof]))
        owlready_ce = onto._onto.charge.only(owlready2.ConstrainedDatatype(float, max_inclusive=1.2) &
                                             owlready2.OneOf([2.3, 5.9, 7.2]))
        self.assertEqual(owlready_ce, to_owlready.map_concept(ce))

        ce = OWLDataSomeValuesFrom(charge, OWLDataIntersectionOf([res, BooleanOWLDatatype]))
        owlready_ce = onto._onto.charge.some(
            owlready2.And([owlready2.ConstrainedDatatype(float, max_inclusive=1.2), bool]))
        self.assertEqual(owlready_ce, to_owlready.map_concept(ce))

        ce = OWLDataMinCardinality(2, act, res)
        owlready_ce = onto._onto.act.min(2, owlready2.ConstrainedDatatype(float, max_inclusive=1.2))
        self.assertEqual(owlready_ce, to_owlready.map_concept(ce))

        ce = OWLDataMaxCardinality(4, charge, DurationOWLDatatype)
        owlready_ce = onto._onto.charge.max(4, Timedelta)
        self.assertEqual(owlready_ce, to_owlready.map_concept(ce))

        ce = OWLDataExactCardinality(3, charge, OWLDataComplementOf(IntegerOWLDatatype))
        owlready_ce = onto._onto.charge.exactly(3, owlready2.Not(int))
        self.assertEqual(owlready_ce, to_owlready.map_concept(ce))

    def test_mapping_rev(self):
        NS = "http://example.com/father#"
        mgr = OWLOntologyManager_Owlready2()
        onto = mgr.load_ontology(IRI.create("file://KGs/father.owl"))

        male = onto._onto.male
        female = onto._onto.female
        has_child = onto._onto.hasChild

        from_owlready = owlapy.owlready2.utils.FromOwlready2()

        ce = male | female
        owl_ce = OWLObjectUnionOf((OWLClass(IRI.create(NS, 'male')), OWLClass(IRI.create(NS, 'female'))))
        self.assertEqual(owl_ce, from_owlready.map_concept(ce))

        ce = male & female
        owl_ce = OWLObjectIntersectionOf((OWLClass(IRI.create(NS, 'male')), OWLClass(IRI.create(NS, 'female'))))
        self.assertEqual(owl_ce, from_owlready.map_concept(ce))

        ce = has_child.some(owlready2.owl.Thing)
        owl_ce = OWLObjectSomeValuesFrom(OWLObjectProperty(IRI(NS, 'hasChild')), OWLThing)
        self.assertEqual(owl_ce, from_owlready.map_concept(ce))

        ce = has_child.only(male)
        owl_ce = OWLObjectAllValuesFrom(OWLObjectProperty(IRI(NS, 'hasChild')), OWLClass(IRI.create(NS, 'male')))
        self.assertEqual(owl_ce, from_owlready.map_concept(ce))

        ce = owlready2.Not(male)
        owl_ce = OWLObjectComplementOf(OWLClass(IRI(NS, 'male')))
        self.assertEqual(owl_ce, from_owlready.map_concept(ce))

        ce = owlready2.OneOf([onto._onto.markus, onto._onto.anna])
        owl_ce = OWLObjectOneOf([OWLNamedIndividual(IRI.create(NS, 'markus')),
                                 OWLNamedIndividual(IRI.create(NS, 'anna'))])
        self.assertEqual(owl_ce, from_owlready.map_concept(ce))

        ce = onto._onto.hasChild.min(2, onto._onto.male)
        owl_ce = OWLObjectMinCardinality(2, OWLObjectProperty(IRI(NS, 'hasChild')), OWLClass(IRI(NS, 'male')))
        self.assertEqual(owl_ce, from_owlready.map_concept(ce))

        ce = onto._onto.hasChild.max(5, onto._onto.female)
        owl_ce = OWLObjectMaxCardinality(5, OWLObjectProperty(IRI(NS, 'hasChild')), OWLClass(IRI(NS, 'female')))
        self.assertEqual(owl_ce, from_owlready.map_concept(ce))

        ce = onto._onto.hasChild.exactly(3, owlready2.owl.Thing)
        owl_ce = OWLObjectExactCardinality(3, OWLObjectProperty(IRI(NS, 'hasChild')), OWLThing)
        self.assertEqual(owl_ce, from_owlready.map_concept(ce))

        ce = has_child.value(onto._onto.markus)
        owl_ce = OWLObjectHasValue(OWLObjectProperty(IRI(NS, 'hasChild')), OWLNamedIndividual(IRI.create(NS, 'markus')))
        self.assertEqual(owl_ce, from_owlready.map_concept(ce))

    def test_mapping_rev_data_properties(self):
        NS = "http://dl-learner.org/mutagenesis#"
        mgr = OWLOntologyManager_Owlready2()
        onto = mgr.load_ontology(IRI.create("file://KGs/Mutagenesis/mutagenesis.owl"))

        act = onto._onto.act
        charge = onto._onto.charge

        from_owlready = owlapy.owlready2.utils.FromOwlready2()

        ce = act.some(float)
        owl_ce = OWLDataSomeValuesFrom(OWLDataProperty(IRI(NS, 'act')), DoubleOWLDatatype)
        self.assertEqual(owl_ce, from_owlready.map_concept(ce))

        ce = charge.only(owlready2.Not(owlready2.ConstrainedDatatype(int, max_inclusive=2)))
        res = OWLDatatypeMaxInclusiveRestriction(2)
        owl_ce = OWLDataAllValuesFrom(OWLDataProperty(IRI(NS, 'charge')), OWLDataComplementOf(res))
        self.assertEqual(owl_ce, from_owlready.map_concept(ce))

        ce = charge.some(owlready2.Not(owlready2.ConstrainedDatatype(int, max_inclusive=2)) |
                         owlready2.ConstrainedDatatype(float, min_inclusive=2.1, max_inclusive=2.2))
        res = OWLDatatypeMaxInclusiveRestriction(2)
        res2 = OWLDatatypeMinMaxInclusiveRestriction(2.1, 2.2)
        owl_ce = OWLDataSomeValuesFrom(OWLDataProperty(IRI(NS, 'charge')),
                                       OWLDataUnionOf([OWLDataComplementOf(res), res2]))
        self.assertEqual(owl_ce, from_owlready.map_concept(ce))

        ce = act.only(owlready2.Not(owlready2.ConstrainedDatatype(int, max_inclusive=2)) & datetime)
        owl_ce = OWLDataAllValuesFrom(OWLDataProperty(IRI(NS, 'act')),
                                      OWLDataIntersectionOf([OWLDataComplementOf(res), DateTimeOWLDatatype]))
        self.assertEqual(owl_ce, from_owlready.map_concept(ce))

        ce = act.some(owlready2.Not(owlready2.OneOf([1, 2, 3])) & Timedelta)
        values = OWLDataOneOf([OWLLiteral(1), OWLLiteral(2), OWLLiteral(3)])
        owl_ce = OWLDataSomeValuesFrom(OWLDataProperty(IRI(NS, 'act')),
                                       OWLDataIntersectionOf([OWLDataComplementOf(values), DurationOWLDatatype]))
        self.assertEqual(owl_ce, from_owlready.map_concept(ce))

        ce = onto._onto.act.value(19.5)
        owl_ce = OWLDataHasValue(OWLDataProperty(IRI(NS, 'act')), OWLLiteral(19.5))
        self.assertEqual(owl_ce, from_owlready.map_concept(ce))

        ce = onto._onto.act.min(2, owlready2.ConstrainedDatatype(int, max_inclusive=2))
        owl_ce = OWLDataMinCardinality(2, OWLDataProperty(IRI(NS, 'act')), res)
        self.assertEqual(owl_ce, from_owlready.map_concept(ce))

        ce = onto._onto.charge.max(5, date)
        owl_ce = OWLDataMaxCardinality(5, OWLDataProperty(IRI(NS, 'charge')), DateOWLDatatype)
        self.assertEqual(owl_ce, from_owlready.map_concept(ce))

        ce = onto._onto.charge.exactly(3, owlready2.Not(float))
        owl_ce = OWLDataExactCardinality(3, OWLDataProperty(IRI(NS, 'charge')), OWLDataComplementOf(DoubleOWLDatatype))
        self.assertEqual(owl_ce, from_owlready.map_concept(ce))


class Owlapy_Owlready2_TempClasses_Test(unittest.TestCase):
    # noinspection DuplicatedCode
    def test_instances(self):
        NS = "http://example.com/father#"
        mgr = OWLOntologyManager_Owlready2()
        onto = mgr.load_ontology(IRI.create("file://KGs/father.owl"))

        male = OWLClass(IRI.create(NS, 'male'))
        female = OWLClass(IRI.create(NS, 'female'))
        has_child = OWLObjectProperty(IRI(NS, 'hasChild'))

        # reasoner = OWLReasoner_Owlready2(onto)
        reasoner = OWLReasoner_Owlready2_TempClasses(onto)

        inst = frozenset(reasoner.instances(female))
        target_inst = frozenset({OWLNamedIndividual(IRI(NS, 'anna')),
                                 OWLNamedIndividual(IRI(NS, 'michelle'))})
        self.assertEqual(inst, target_inst)

        inst = frozenset(reasoner.instances(
            OWLObjectIntersectionOf((male, OWLObjectSomeValuesFrom(property=has_child, filler=female)))))
        target_inst = frozenset({OWLNamedIndividual(IRI(NS, 'markus'))})
        self.assertEqual(inst, target_inst)

        inst = frozenset(reasoner.instances(
            OWLObjectIntersectionOf((female, OWLObjectSomeValuesFrom(property=has_child, filler=OWLThing)))))
        target_inst = frozenset({OWLNamedIndividual(IRI(NS, 'anna'))})
        self.assertEqual(inst, target_inst)


if __name__ == '__main__':
    unittest.main()



#  This file is part of the OWL API.
#  * The contents of this file are subject to the LGPL License, Version 3.0.
#  * Copyright 2014, The University of Manchester
#
# * This program is free software: you can redistribute it and/or modify it under the terms of the GNU General Public
# License as published by the Free Software Foundation, either version 3 of the License, or (at your option) any
# later version. * This program is distributed in the hope that it will be useful, but WITHOUT ANY WARRANTY; without
# even the implied warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public
# License for more details. * You should have received a copy of the GNU General Public License along with this
# program.  If not, see http://www.gnu.org/licenses/.
#
# * Alternatively, the contents of this file may be used under the terms of the Apache License, Version 2.0 in which
# case, the provisions of the Apache License Version 2.0 are applicable instead of those above. * Licensed under the
# Apache License, Version 2.0 (the "License"); you may not use this file except in compliance with the License. You
# may obtain a copy of the License at * http://www.apache.org/licenses/LICENSE-2.0 * Unless required by applicable
# law or agreed to in writing, software distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the License for the specific language
# governing permissions and limitations under the License.
#
# package: org.semanticweb.owlapi.api.test.axioms
#
#  * @author Matthew Horridge, The University of Manchester, Information Management Group
#  * @since 3.0.0
#
import unittest

from owlapy.model import OWLObjectProperty, OWLNamedIndividual, OWLObjectComplementOf, \
    OWLObjectAllValuesFrom, OWLObjectSomeValuesFrom, OWLObjectIntersectionOf, OWLObjectUnionOf, \
    OWLObjectMinCardinality, OWLObjectMaxCardinality, OWLObjectHasValue, OWLObjectOneOf, OWLClassExpression, IRI, \
    BooleanOWLDatatype, DoubleOWLDatatype, IntegerOWLDatatype, OWLClass, OWLDataAllValuesFrom, OWLDataComplementOf, \
    OWLDataIntersectionOf, OWLDataProperty, OWLDataSomeValuesFrom, OWLDataUnionOf, \
    OWLDataHasValue, OWLDataMaxCardinality, OWLDataMinCardinality, OWLDataOneOf, OWLLiteral
from owlapy.model.providers import OWLDatatypeMinMaxExclusiveRestriction
from owlapy.util import NNF


def iri(suffix):
    NS = "http://example.org/"
    return IRI.create(NS, suffix)


class Owlapy_NNF_Test(unittest.TestCase):
    """ generated source for class NNFTestCase """
    clsA = OWLClass(iri("A"))
    clsB = OWLClass(iri("B"))
    clsC = OWLClass(iri("C"))
    clsD = OWLClass(iri("D"))
    propP = OWLObjectProperty(iri("p"))
    indA = OWLNamedIndividual(iri("a"))

    def get_nnf(self, ce: OWLClassExpression):
        return NNF().get_class_nnf(ce)

    def testPosOWLClass(self):
        """ generated source for method testPosOWLClass """
        cls = OWLClass(iri("A"))
        self.assertEqual(cls.get_nnf(), cls)

    def testNegOWLClass(self):
        """ generated source for method testNegOWLClass """
        cls = OWLObjectComplementOf(OWLClass(iri("A")))
        self.assertEqual(cls.get_nnf(), cls)

    def testPosAllValuesFrom(self):
        """ generated source for method testPosAllValuesFrom """
        cls = OWLObjectAllValuesFrom(OWLObjectProperty(iri("p")), OWLClass(iri("A")))
        self.assertEqual(cls.get_nnf(), cls)

    def testNegAllValuesFrom(self):
        """ generated source for method testNegAllValuesFrom """
        property = OWLObjectProperty(iri("p"))
        filler = OWLClass(iri("A"))
        all_values_from = OWLObjectAllValuesFrom(property, filler)
        cls = all_values_from.get_object_complement_of()
        nnf = OWLObjectSomeValuesFrom(property, filler.get_object_complement_of())
        self.assertEqual(cls.get_nnf(), nnf)

    def testPosSomeValuesFrom(self):
        """ generated source for method testPosSomeValuesFrom """
        cls = OWLObjectSomeValuesFrom(OWLObjectProperty(iri("p")), OWLClass(iri("A")))
        self.assertEqual(cls.get_nnf(), cls)

    def testNegSomeValuesFrom(self):
        """ generated source for method testNegSomeValuesFrom """
        property = OWLObjectProperty(iri("p"))
        filler = OWLClass(iri("A"))
        some_values_from = OWLObjectSomeValuesFrom(property, filler)
        cls = OWLObjectComplementOf(some_values_from)
        nnf = OWLObjectAllValuesFrom(property, OWLObjectComplementOf(filler))
        self.assertEqual(cls.get_nnf(), nnf)

    def testPosObjectIntersectionOf(self):
        """ generated source for method testPosObjectIntersectionOf """
        cls = OWLObjectIntersectionOf((OWLClass(iri("A")), OWLClass(iri("B")), OWLClass(iri("C"))))
        self.assertEqual(cls.get_nnf(), cls)

    def testNegObjectIntersectionOf(self):
        """ generated source for method testNegObjectIntersectionOf """
        cls = OWLObjectComplementOf(OWLObjectIntersectionOf(
            (OWLClass(iri("A")), OWLClass(iri("B")), OWLClass(iri("C")))))
        nnf = OWLObjectUnionOf(
            (OWLObjectComplementOf(OWLClass(iri("A"))),
             OWLObjectComplementOf(OWLClass(iri("B"))),
             OWLObjectComplementOf(OWLClass(iri("C")))))
        self.assertEqual(cls.get_nnf(), nnf)

    def testPosObjectUnionOf(self):
        """ generated source for method testPosObjectUnionOf """
        cls = OWLObjectUnionOf((OWLClass(iri("A")), OWLClass(iri("B")), OWLClass(iri("C"))))
        self.assertEqual(cls.get_nnf(), cls)

    def testNegObjectUnionOf(self):
        """ generated source for method testNegObjectUnionOf """
        cls = OWLObjectComplementOf(OWLObjectUnionOf((OWLClass(iri("A")), OWLClass(iri("B")), OWLClass(iri("C")))))
        nnf = OWLObjectIntersectionOf(
            (OWLObjectComplementOf(OWLClass(iri("A"))),
             OWLObjectComplementOf(OWLClass(iri("B"))),
             OWLObjectComplementOf(OWLClass(iri("C")))))
        self.assertEqual(cls.get_nnf(), nnf)

    def testPosObjectMinCardinality(self):
        """ generated source for method testPosObjectMinCardinality """
        prop = OWLObjectProperty(iri("p"))
        filler = OWLClass(iri("A"))
        cls = OWLObjectMinCardinality(cardinality=3, property=prop, filler=filler)
        self.assertEqual(cls.get_nnf(), cls)

    def testNegObjectMinCardinality(self):
        """ generated source for method testNegObjectMinCardinality """
        prop = OWLObjectProperty(iri("p"))
        filler = OWLClass(iri("A"))
        cls = OWLObjectMinCardinality(cardinality=3, property=prop, filler=filler).get_object_complement_of()
        nnf = OWLObjectMaxCardinality(cardinality=2, property=prop, filler=filler)
        self.assertEqual(cls.get_nnf(), nnf)

    def testPosObjectMaxCardinality(self):
        """ generated source for method testPosObjectMaxCardinality """
        prop = OWLObjectProperty(iri("p"))
        filler = OWLClass(iri("A"))
        cls = OWLObjectMaxCardinality(cardinality=3, property=prop, filler=filler)
        self.assertEqual(cls.get_nnf(), cls)

    def testNegObjectMaxCardinality(self):
        """ generated source for method testNegObjectMaxCardinality """
        prop = OWLObjectProperty(iri("p"))
        filler = OWLClass(iri("A"))
        cls = OWLObjectMaxCardinality(cardinality=3, property=prop, filler=filler).get_object_complement_of()
        nnf = OWLObjectMinCardinality(cardinality=4, property=prop, filler=filler)
        self.assertEqual(cls.get_nnf(), nnf)

    def testNamedClass(self):
        """ generated source for method testNamedClass """
        desc = self.clsA
        nnf = self.clsA
        comp = self.get_nnf(desc)
        self.assertEqual(nnf, comp)

    def testObjectIntersectionOf(self):
        """ generated source for method testObjectIntersectionOf """
        desc = OWLObjectIntersectionOf((self.clsA, self.clsB))
        neg = OWLObjectComplementOf(desc)
        nnf = OWLObjectUnionOf((OWLObjectComplementOf(self.clsA), OWLObjectComplementOf(self.clsB)))
        comp = self.get_nnf(neg)
        self.assertEqual(nnf, comp)

    def testObjectUnionOf(self):
        """ generated source for method testObjectUnionOf """
        desc = OWLObjectUnionOf((self.clsA, self.clsB))
        neg = OWLObjectComplementOf(desc)
        nnf = OWLObjectIntersectionOf((OWLObjectComplementOf(self.clsA), OWLObjectComplementOf(self.clsB)))
        comp = self.get_nnf(neg)
        self.assertEqual(nnf, comp)

    def testDoubleNegation(self):
        """ generated source for method testDoubleNegation """
        desc = OWLObjectComplementOf(self.clsA)
        neg = OWLObjectComplementOf(desc)
        nnf = self.clsA
        comp = self.get_nnf(neg)
        self.assertEqual(nnf, comp)

    def testTripleNegation(self):
        """ generated source for method testTripleNegation """
        desc = OWLObjectComplementOf(OWLObjectComplementOf(self.clsA))
        neg = OWLObjectComplementOf(desc)
        nnf = OWLObjectComplementOf(self.clsA)
        comp = self.get_nnf(neg)
        self.assertEqual(nnf, comp)

    def testObjectSome(self):
        """ generated source for method testObjectSome """
        desc = OWLObjectSomeValuesFrom(self.propP, self.clsA)
        neg = OWLObjectComplementOf(desc)
        nnf = OWLObjectAllValuesFrom(self.propP, OWLObjectComplementOf(self.clsA))
        comp = self.get_nnf(neg)
        self.assertEqual(nnf, comp)

    def testObjectAll(self):
        """ generated source for method testObjectAll """
        desc = OWLObjectAllValuesFrom(self.propP, self.clsA)
        neg = OWLObjectComplementOf(desc)
        nnf = OWLObjectSomeValuesFrom(self.propP, OWLObjectComplementOf(self.clsA))
        comp = self.get_nnf(neg)
        self.assertEqual(nnf, comp)

    def testObjectHasValue(self):
        """ generated source for method testObjectHasValue """
        desc = OWLObjectHasValue(self.propP, self.indA)
        neg = OWLObjectComplementOf(desc)
        nnf = OWLObjectAllValuesFrom(self.propP, OWLObjectComplementOf(OWLObjectOneOf(self.indA)))
        comp = self.get_nnf(neg)
        self.assertEqual(nnf, comp)

    def testObjectMin(self):
        """ generated source for method testObjectMin """
        desc = OWLObjectMinCardinality(cardinality=3, property=self.propP, filler=self.clsA)
        neg = OWLObjectComplementOf(desc)
        nnf = OWLObjectMaxCardinality(cardinality=2, property=self.propP, filler=self.clsA)
        comp = self.get_nnf(neg)
        self.assertEqual(nnf, comp)

    def testObjectMax(self):
        """ generated source for method testObjectMax """
        desc = OWLObjectMaxCardinality(cardinality=3, property=self.propP, filler=self.clsA)
        neg = OWLObjectComplementOf(desc)
        nnf = OWLObjectMinCardinality(cardinality=4, property=self.propP, filler=self.clsA)
        comp = self.get_nnf(neg)
        self.assertEqual(nnf, comp)

    def testNestedA(self):
        """ generated source for method testNestedA """
        filler_a = OWLObjectUnionOf((self.clsA, self.clsB))
        op_a = OWLObjectSomeValuesFrom(self.propP, filler_a)
        op_b = self.clsB
        desc = OWLObjectUnionOf((op_a, op_b))
        nnf = OWLObjectIntersectionOf(
            (OWLObjectComplementOf(self.clsB),
             OWLObjectAllValuesFrom(self.propP,
                                    OWLObjectIntersectionOf((OWLObjectComplementOf(self.clsA),
                                                             OWLObjectComplementOf(self.clsB))))))
        neg = OWLObjectComplementOf(desc)
        comp = self.get_nnf(neg)
        self.assertEqual(comp, nnf)

    def testNestedB(self):
        """ generated source for method testNestedB """
        desc = OWLObjectIntersectionOf(
            (OWLObjectIntersectionOf((self.clsA, self.clsB)),
             OWLObjectComplementOf(OWLObjectUnionOf((self.clsC, self.clsD)))))
        neg = OWLObjectComplementOf(desc)
        nnf = OWLObjectUnionOf(
            (OWLObjectUnionOf((OWLObjectComplementOf(self.clsA),
                               OWLObjectComplementOf(self.clsB))),
             OWLObjectUnionOf((self.clsC, self.clsD))))
        comp = self.get_nnf(neg)
        self.assertEqual(comp, nnf)

    def testPosDataAllValuesFrom(self):
        cls = OWLDataAllValuesFrom(OWLDataProperty(iri("p")), IntegerOWLDatatype)
        self.assertEqual(cls.get_nnf(), cls)

    def testNegDataAllValuesFrom(self):
        property = OWLDataProperty(iri("p"))
        all_values_from = OWLDataAllValuesFrom(property, IntegerOWLDatatype)
        cls = all_values_from.get_object_complement_of()
        nnf = OWLDataSomeValuesFrom(property, OWLDataComplementOf(IntegerOWLDatatype))
        self.assertEqual(cls.get_nnf(), nnf)

    def testPosDataSomeValuesFrom(self):
        cls = OWLDataSomeValuesFrom(OWLDataProperty(iri("p")), IntegerOWLDatatype)
        self.assertEqual(cls.get_nnf(), cls)

    def testNegDataSomeValuesFrom(self):
        property = OWLDataProperty(iri("p"))
        some_values_from = OWLDataSomeValuesFrom(property, IntegerOWLDatatype)
        cls = OWLDataComplementOf(some_values_from)
        nnf = OWLDataAllValuesFrom(property, OWLDataComplementOf(IntegerOWLDatatype))
        self.assertEqual(self.get_nnf(cls), nnf)

    def testPosDataIntersectionOf(self):
        cls = OWLDataIntersectionOf((BooleanOWLDatatype, DoubleOWLDatatype, IntegerOWLDatatype))
        self.assertEqual(self.get_nnf(cls), cls)

    def testNegDataIntersectionOf(self):
        cls = OWLDataComplementOf(OWLDataIntersectionOf(
            (BooleanOWLDatatype, DoubleOWLDatatype, IntegerOWLDatatype)))
        nnf = OWLDataUnionOf(
            (OWLDataComplementOf(BooleanOWLDatatype),
             OWLDataComplementOf(DoubleOWLDatatype),
             OWLDataComplementOf(IntegerOWLDatatype)))
        self.assertEqual(self.get_nnf(cls), nnf)

    def testPosDataUnionOf(self):
        cls = OWLDataUnionOf((BooleanOWLDatatype, DoubleOWLDatatype, IntegerOWLDatatype))
        self.assertEqual(self.get_nnf(cls), cls)

    def testNegDataUnionOf(self):
        cls = OWLDataComplementOf(OWLDataUnionOf((BooleanOWLDatatype, DoubleOWLDatatype, IntegerOWLDatatype)))
        nnf = OWLDataIntersectionOf(
            (OWLDataComplementOf(BooleanOWLDatatype),
             OWLDataComplementOf(DoubleOWLDatatype),
             OWLDataComplementOf(IntegerOWLDatatype)))
        self.assertEqual(self.get_nnf(cls), nnf)

    def testPosDataMinCardinality(self):
        prop = OWLDataProperty(iri("p"))
        cls = OWLDataMinCardinality(cardinality=3, property=prop, filler=IntegerOWLDatatype)
        self.assertEqual(cls.get_nnf(), cls)

    def testNegDataMinCardinality(self):
        prop = OWLDataProperty(iri("p"))
        filler = IntegerOWLDatatype
        cls = OWLDataMinCardinality(cardinality=3, property=prop, filler=filler).get_object_complement_of()
        nnf = OWLDataMaxCardinality(cardinality=2, property=prop, filler=filler)
        self.assertEqual(cls.get_nnf(), nnf)

    def testPosDataMaxCardinality(self):
        prop = OWLDataProperty(iri("p"))
        cls = OWLDataMaxCardinality(cardinality=3, property=prop, filler=IntegerOWLDatatype)
        self.assertEqual(cls.get_nnf(), cls)

    def testNegDataMaxCardinality(self):
        prop = OWLDataProperty(iri("p"))
        filler = IntegerOWLDatatype
        cls = OWLDataMaxCardinality(cardinality=3, property=prop, filler=filler).get_object_complement_of()
        nnf = OWLDataMinCardinality(cardinality=4, property=prop, filler=filler)
        self.assertEqual(cls.get_nnf(), nnf)

    def testDatatype(self):
        desc = IntegerOWLDatatype
        nnf = IntegerOWLDatatype
        comp = self.get_nnf(desc)
        self.assertEqual(nnf, comp)

    def testDataDoubleNegation(self):
        desc = OWLDataComplementOf(IntegerOWLDatatype)
        neg = OWLDataComplementOf(desc)
        nnf = IntegerOWLDatatype
        comp = self.get_nnf(neg)
        self.assertEqual(nnf, comp)

    def testDataTripleNegation(self):
        desc = OWLDataComplementOf(OWLDataComplementOf(IntegerOWLDatatype))
        neg = OWLDataComplementOf(desc)
        nnf = OWLDataComplementOf(IntegerOWLDatatype)
        comp = self.get_nnf(neg)
        self.assertEqual(nnf, comp)

    def testDataHasValue(self):
        prop = OWLDataProperty(iri("p"))
        literal = OWLLiteral(5)
        desc = OWLDataHasValue(prop, literal)
        neg = OWLDataComplementOf(desc)
        nnf = OWLDataAllValuesFrom(prop, OWLDataComplementOf(OWLDataOneOf(literal)))
        comp = self.get_nnf(neg)
        self.assertEqual(nnf, comp)

    def testDataNestedA(self):
        restriction = OWLDatatypeMinMaxExclusiveRestriction(5, 6)
        prop = OWLDataProperty(iri("p"))
        filler_a = OWLDataUnionOf((IntegerOWLDatatype, DoubleOWLDatatype))
        op_a = OWLDataSomeValuesFrom(prop, filler_a)
        op_b = OWLDataIntersectionOf((restriction, IntegerOWLDatatype))
        desc = OWLDataUnionOf((op_a, op_b))
        nnf = OWLDataIntersectionOf(
                (OWLDataAllValuesFrom(prop, OWLDataIntersectionOf((OWLDataComplementOf(DoubleOWLDatatype),
                                                                   OWLDataComplementOf(IntegerOWLDatatype)))),
                 OWLDataUnionOf((OWLDataComplementOf(IntegerOWLDatatype), OWLDataComplementOf(restriction)))))
        neg = OWLDataComplementOf(desc)
        comp = self.get_nnf(neg)
        self.assertEqual(comp, nnf)

    def testDataNestedB(self):
        desc = OWLDataIntersectionOf(
            (OWLDataIntersectionOf((IntegerOWLDatatype, DoubleOWLDatatype)),
             OWLDataComplementOf(OWLDataUnionOf((BooleanOWLDatatype, OWLDataOneOf(OWLLiteral(True)))))))
        neg = OWLDataComplementOf(desc)
        nnf = OWLDataUnionOf(
            (OWLDataUnionOf((BooleanOWLDatatype, OWLDataOneOf(OWLLiteral(True)))),
             OWLDataUnionOf((OWLDataComplementOf(DoubleOWLDatatype),
                             OWLDataComplementOf(IntegerOWLDatatype)))))
        comp = self.get_nnf(neg)
        self.assertEqual(comp, nnf)



import os

from experiments_standard import ClosedWorld_ReasonerFactory
from ontolearn.knowledge_base import KnowledgeBase
from ontolearn.learning_problem_generator import LearningProblemGenerator
from ontolearn.utils import setup_logging

setup_logging("logging_test.conf")

try:
    os.chdir("examples")
except FileNotFoundError:
    pass

path = '../KGs/Biopax/biopax.owl'

# kb = KnowledgeBase(path=path, reasoner_factory=OWLReasoner_Owlready2_TempClasses)
kb = KnowledgeBase(path=path, reasoner_factory=ClosedWorld_ReasonerFactory)
lp = LearningProblemGenerator(knowledge_base=kb)
num_inds = kb.individuals_count()
concepts = list(lp.get_concepts(num_problems=5000,
                                num_diff_runs=10,
                                min_num_instances=int(2),
                                max_num_instances=int(num_inds * .95),
                                min_length=4, max_length=40))
# Each generated concept defines the type information of min 10% and max 80% of instances.
# for c in concepts:
#     print('*', c)

lp.export_concepts(concepts, path='example_concepts')



from collections import Counter
from functools import singledispatchmethod
from typing import Iterable, Generic, TypeVar, Callable, List

from owlapy.model import OWLDataProperty, OWLObject, OWLClass, OWLObjectProperty, OWLObjectSomeValuesFrom, \
    OWLObjectAllValuesFrom, OWLObjectUnionOf, OWLObjectIntersectionOf, OWLObjectComplementOf, OWLObjectInverseOf, \
    OWLObjectCardinalityRestriction, OWLObjectHasSelf, OWLObjectHasValue, OWLObjectOneOf, OWLNamedIndividual, \
    OWLObjectMinCardinality, OWLObjectExactCardinality, OWLObjectMaxCardinality, OWLClassExpression, OWLThing, \
    OWLDataSomeValuesFrom, OWLDataOneOf, OWLDatatypeRestriction, OWLDataComplementOf, OWLDataAllValuesFrom, \
    OWLDataCardinalityRestriction, OWLDatatype, OWLDataHasValue, OWLDataUnionOf, OWLDataIntersectionOf \

from owlapy.util import OrderedOWLObject, iter_count
from sortedcontainers import SortedSet


class OWLClassExpressionLengthMetric:
    """Length calculation of OWLClassExpression

    Args:
        class_length: Class: "C"
        object_intersection_length: Intersection: A ⨅ B
        object_union_length: Union: A ⨆ B
        object_complement_length: Complement: ¬ C
        object_some_values_length: Obj. Some Values: ∃ r.C
        object_all_values_length: Obj. All Values: ∀ r.C
        object_has_value_length: Obj. Has Value: ∃ r.{I}
        object_cardinality_length: Obj. Cardinality restriction: ≤n r.C
        object_has_self_length: Obj. Self restriction: ∃ r.Self
        object_one_of_length: Obj. One of: ∃ r.{X,Y,Z}
        data_some_values_length: Data Some Values: ∃ p.t
        data_all_values_length: Data All Values: ∀ p.t
        data_has_value_length: Data Has Value: ∃ p.{V}
        data_cardinality_length: Data Cardinality restriction: ≤n r.t
        object_property_length: Obj. Property: ∃ r.C
        object_inverse_length: Inverse property: ∃ r⁻.C
        data_property_length: Data Property: ∃ p.t
        datatype_length: Datatype: ^^datatype
        data_one_of_length: Data One of: ∃ p.{U,V,W}
        data_complement_length: Data Complement: ¬datatype
        data_intersection_length: Data Intersection: datatype ⨅ datatype
        data_union_length: Data Union: datatype ⨆ datatype
    """

    __slots__ = 'class_length', 'object_intersection_length', 'object_union_length', 'object_complement_length', \
                'object_some_values_length', 'object_all_values_length', 'object_has_value_length', \
                'object_cardinality_length', 'object_has_self_length', 'object_one_of_length', \
                'data_some_values_length', 'data_all_values_length', 'data_has_value_length', \
                'data_cardinality_length', 'object_property_length', 'object_inverse_length', 'data_property_length', \
                'datatype_length', 'data_one_of_length', 'data_complement_length', 'data_intersection_length', \
                'data_union_length'

    class_length: int
    object_intersection_length: int
    object_union_length: int
    object_complement_length: int
    object_some_values_length: int
    object_all_values_length: int
    object_has_value_length: int
    object_cardinality_length: int
    object_has_self_length: int
    object_one_of_length: int
    data_some_values_length: int
    data_all_values_length: int
    data_has_value_length: int
    data_cardinality_length: int
    object_property_length: int
    object_inverse_length: int
    data_property_length: int
    datatype_length: int
    data_one_of_length: int
    data_complement_length: int
    data_intersection_length: int
    data_union_length: int

    def __init__(self, *,
                 class_length: int,
                 object_intersection_length: int,
                 object_union_length: int,
                 object_complement_length: int,
                 object_some_values_length: int,
                 object_all_values_length: int,
                 object_has_value_length: int,
                 object_cardinality_length: int,
                 object_has_self_length: int,
                 object_one_of_length: int,
                 data_some_values_length: int,
                 data_all_values_length: int,
                 data_has_value_length: int,
                 data_cardinality_length: int,
                 object_property_length: int,
                 object_inverse_length: int,
                 data_property_length: int,
                 datatype_length: int,
                 data_one_of_length: int,
                 data_complement_length: int,
                 data_intersection_length: int,
                 data_union_length: int,
                 ):
        self.class_length = class_length
        self.object_intersection_length = object_intersection_length
        self.object_union_length = object_union_length
        self.object_complement_length = object_complement_length
        self.object_some_values_length = object_some_values_length
        self.object_all_values_length = object_all_values_length
        self.object_has_value_length = object_has_value_length
        self.object_cardinality_length = object_cardinality_length
        self.object_has_self_length = object_has_self_length
        self.object_one_of_length = object_one_of_length
        self.data_some_values_length = data_some_values_length
        self.data_all_values_length = data_all_values_length
        self.data_has_value_length = data_has_value_length
        self.data_cardinality_length = data_cardinality_length
        self.object_property_length = object_property_length
        self.object_inverse_length = object_inverse_length
        self.data_property_length = data_property_length
        self.datatype_length = datatype_length
        self.data_one_of_length = data_one_of_length
        self.data_complement_length = data_complement_length
        self.data_intersection_length = data_intersection_length
        self.data_union_length = data_union_length

    @staticmethod
    def get_default() -> 'OWLClassExpressionLengthMetric':
        return OWLClassExpressionLengthMetric(
            class_length=1,
            object_intersection_length=1,
            object_union_length=1,
            object_complement_length=1,
            object_some_values_length=1,
            object_all_values_length=1,
            object_has_value_length=2,
            object_cardinality_length=2,
            object_has_self_length=1,
            object_one_of_length=1,
            data_some_values_length=1,
            data_all_values_length=1,
            data_has_value_length=2,
            data_cardinality_length=2,
            object_property_length=1,
            object_inverse_length=2,
            data_property_length=1,
            datatype_length=1,
            data_one_of_length=1,
            data_complement_length=1,
            data_intersection_length=1,
            data_union_length=1,
        )

    # single dispatch is still not implemented in mypy, see https://github.com/python/mypy/issues/2904
    @singledispatchmethod
    def length(self, o: OWLObject) -> int:
        raise NotImplementedError

    @length.register
    def _(self, o: OWLClass) -> int:
        return self.class_length

    @length.register
    def _(self, p: OWLObjectProperty) -> int:
        return self.object_property_length

    @length.register
    def _(self, e: OWLObjectSomeValuesFrom) -> int:
        return self.object_some_values_length \
               + self.length(e.get_property()) \
               + self.length(e.get_filler())

    @length.register
    def _(self, e: OWLObjectAllValuesFrom) -> int:
        return self.object_all_values_length \
               + self.length(e.get_property()) \
               + self.length(e.get_filler())

    @length.register
    def _(self, c: OWLObjectUnionOf) -> int:
        length = -self.object_union_length
        for op in c.operands():
            length += self.length(op) + self.object_union_length

        return length

    @length.register
    def _(self, c: OWLObjectIntersectionOf) -> int:
        length = -self.object_intersection_length
        for op in c.operands():
            length += self.length(op) + self.object_intersection_length

        return length

    @length.register
    def _(self, n: OWLObjectComplementOf) -> int:
        return self.length(n.get_operand()) + self.object_complement_length

    @length.register
    def _(self, p: OWLObjectInverseOf) -> int:
        return self.object_inverse_length

    @length.register
    def _(self, e: OWLObjectCardinalityRestriction) -> int:
        return self.object_cardinality_length \
               + self.length(e.get_property()) \
               + self.length(e.get_filler())

    @length.register
    def _(self, s: OWLObjectHasSelf) -> int:
        return self.object_has_self_length + self.length(s.get_property())

    @length.register
    def _(self, v: OWLObjectHasValue) -> int:
        return self.object_has_value_length + self.length(v.get_property())

    @length.register
    def _(self, o: OWLObjectOneOf) -> int:
        return self.object_one_of_length

    @length.register
    def _(self, p: OWLDataProperty) -> int:
        return self.data_property_length

    @length.register
    def _(self, e: OWLDataSomeValuesFrom) -> int:
        return self.data_some_values_length \
               + self.length(e.get_property()) \
               + self.length(e.get_filler())

    @length.register
    def _(self, e: OWLDataAllValuesFrom) -> int:
        return self.data_all_values_length \
               + self.length(e.get_property()) \
               + self.length(e.get_filler())

    @length.register
    def _(self, e: OWLDataCardinalityRestriction) -> int:
        return self.data_cardinality_length \
               + self.length(e.get_property()) \
               + self.length(e.get_filler())

    @length.register
    def _(self, v: OWLDataHasValue) -> int:
        return self.data_has_value_length + self.length(v.get_property())

    @length.register
    def _(self, o: OWLDataOneOf) -> int:
        return self.data_one_of_length

    @length.register
    def _(self, n: OWLDatatypeRestriction) -> int:
        return iter_count(n.get_facet_restrictions())

    @length.register
    def _(self, n: OWLDataComplementOf) -> int:
        return self.data_complement_length + self.length(n.get_data_range())

    @length.register
    def _(self, c: OWLDataUnionOf) -> int:
        length = -self.data_union_length
        for op in c.operands():
            length += self.length(op) + self.data_union_length
        return length

    @length.register
    def _(self, c: OWLDataIntersectionOf) -> int:
        length = -self.data_intersection_length
        for op in c.operands():
            length += self.length(op) + self.data_intersection_length
        return length

    @length.register
    def _(self, t: OWLDatatype) -> int:
        return self.datatype_length


_N = TypeVar('_N')  #:
_O = TypeVar('_O')  #:


class EvaluatedDescriptionSet(Generic[_N, _O]):
    __slots__ = 'items', '_max_size', '_Ordering'

    items: 'SortedSet[_N]'
    _max_size: int
    _Ordering: Callable[[_N], _O]

    def __init__(self, ordering: Callable[[_N], _O], max_size: int = 10):
        self._max_size = max_size
        self._Ordering = ordering
        self.items = SortedSet(key=self._Ordering)

    def maybe_add(self, node: _N):
        if len(self.items) == self._max_size:
            worst = self.items[0]
            if self._Ordering(node) > self._Ordering(worst):
                self.items.pop(0)
                self.items.add(node)
                return True
        else:
            self.items.add(node)
            return True
        return False

    def clean(self):
        self.items.clear()

    def worst(self):
        return self.items[0]

    def best(self):
        return self.items[-1]

    def best_quality_value(self) -> float:
        return self.items[-1].quality

    def __iter__(self) -> Iterable[_N]:
        yield from reversed(self.items)


def _avoid_overly_redundand_operands(operands: List[_O], max_count: int = 2) -> List[_O]:
    _max_count = max_count
    r = []
    counts = Counter(operands)
    for op in sorted(operands, key=OrderedOWLObject):
        for _ in range(min(_max_count, counts[op])):
            r.append(op)
    return r


def _sort_by_ordered_owl_object(i: Iterable[_O]) -> Iterable[_O]:
    return sorted(i, key=OrderedOWLObject)


class ConceptOperandSorter:
    # single dispatch is still not implemented in mypy, see https://github.com/python/mypy/issues/2904
    @singledispatchmethod
    def sort(self, o: _O) -> _O:
        raise NotImplementedError

    @sort.register
    def _(self, o: OWLClass) -> OWLClass:
        return o

    @sort.register
    def _(self, p: OWLObjectProperty) -> OWLObjectProperty:
        return p

    @sort.register
    def _(self, i: OWLNamedIndividual) -> OWLNamedIndividual:
        return i

    @sort.register
    def _(self, e: OWLObjectSomeValuesFrom) -> OWLObjectSomeValuesFrom:
        t = OWLObjectSomeValuesFrom(property=e.get_property(), filler=self.sort(e.get_filler()))
        if t == e:
            return e
        else:
            return t

    @sort.register
    def _(self, e: OWLObjectAllValuesFrom) -> OWLObjectAllValuesFrom:
        t = OWLObjectAllValuesFrom(property=e.get_property(), filler=self.sort(e.get_filler()))
        if t == e:
            return e
        else:
            return t

    @sort.register
    def _(self, c: OWLObjectUnionOf) -> OWLObjectUnionOf:
        t = OWLObjectUnionOf(_sort_by_ordered_owl_object(c.operands()))
        if t == c:
            return c
        else:
            return t

    @sort.register
    def _(self, c: OWLObjectIntersectionOf) -> OWLObjectIntersectionOf:
        t = OWLObjectIntersectionOf(_sort_by_ordered_owl_object(c.operands()))
        if t == c:
            return c
        else:
            return t

    @sort.register
    def _(self, n: OWLObjectComplementOf) -> OWLObjectComplementOf:
        return n

    @sort.register
    def _(self, p: OWLObjectInverseOf) -> OWLObjectInverseOf:
        return p

    @sort.register
    def _(self, r: OWLObjectMinCardinality) -> OWLObjectMinCardinality:
        t = OWLObjectMinCardinality(cardinality=r.get_cardinality(), property=r.get_property(),
                                    filler=self.sort(r.get_filler()))
        if t == r:
            return r
        else:
            return t

    @sort.register
    def _(self, r: OWLObjectExactCardinality) -> OWLObjectExactCardinality:
        t = OWLObjectExactCardinality(cardinality=r.get_cardinality(), property=r.get_property(),
                                      filler=self.sort(r.get_filler()))
        if t == r:
            return r
        else:
            return t

    @sort.register
    def _(self, r: OWLObjectMaxCardinality) -> OWLObjectMaxCardinality:
        t = OWLObjectMaxCardinality(cardinality=r.get_cardinality(), property=r.get_property(),
                                    filler=self.sort(r.get_filler()))
        if t == r:
            return r
        else:
            return t

    @sort.register
    def _(self, r: OWLObjectHasSelf) -> OWLObjectHasSelf:
        return r

    @sort.register
    def _(self, r: OWLObjectHasValue):
        return r

    @sort.register
    def _(self, r: OWLObjectOneOf):
        t = OWLObjectOneOf(_sort_by_ordered_owl_object(r.individuals()))
        if t == r:
            return r
        else:
            return t


class OperandSetTransform:
    def simplify(self, o: OWLClassExpression) -> OWLClassExpression:
        return self._simplify(o).get_nnf()

    # single dispatch is still not implemented in mypy, see https://github.com/python/mypy/issues/2904
    @singledispatchmethod
    def _simplify(self, o: _O) -> _O:
        raise NotImplementedError

    @_simplify.register
    def _(self, o: OWLClass) -> OWLClass:
        return o

    @_simplify.register
    def _(self, p: OWLObjectProperty) -> OWLObjectProperty:
        return p

    @_simplify.register
    def _(self, i: OWLNamedIndividual) -> OWLNamedIndividual:
        return i

    @_simplify.register
    def _(self, e: OWLObjectSomeValuesFrom) -> OWLObjectSomeValuesFrom:
        return OWLObjectSomeValuesFrom(property=e.get_property(), filler=self._simplify(e.get_filler()))

    @_simplify.register
    def _(self, e: OWLObjectAllValuesFrom) -> OWLObjectAllValuesFrom:
        return OWLObjectAllValuesFrom(property=e.get_property(), filler=self._simplify(e.get_filler()))

    @_simplify.register
    def _(self, c: OWLObjectUnionOf) -> OWLClassExpression:
        s = set(map(self._simplify, set(c.operands())))
        if OWLThing in s:
            return OWLThing
        elif len(s) == 1:
            return s.pop()
        return OWLObjectUnionOf(_sort_by_ordered_owl_object(s))

    @_simplify.register
    def _(self, c: OWLObjectIntersectionOf) -> OWLClassExpression:
        s = set(map(self._simplify, set(c.operands())))
        s.discard(OWLThing)
        if not s:
            return OWLThing
        elif len(s) == 1:
            return s.pop()
        return OWLObjectIntersectionOf(_sort_by_ordered_owl_object(s))

    @_simplify.register
    def _(self, n: OWLObjectComplementOf) -> OWLObjectComplementOf:
        return n

    @_simplify.register
    def _(self, p: OWLObjectInverseOf) -> OWLObjectInverseOf:
        return p

    @_simplify.register
    def _(self, r: OWLObjectMinCardinality) -> OWLObjectMinCardinality:
        return OWLObjectMinCardinality(cardinality=r.get_cardinality(), property=r.get_property(),
                                       filler=self._simplify(r.get_filler()))

    @_simplify.register
    def _(self, r: OWLObjectExactCardinality) -> OWLObjectExactCardinality:
        return OWLObjectExactCardinality(cardinality=r.get_cardinality(), property=r.get_property(),
                                         filler=self._simplify(r.get_filler()))

    @_simplify.register
    def _(self, r: OWLObjectMaxCardinality) -> OWLObjectMaxCardinality:
        return OWLObjectMaxCardinality(cardinality=r.get_cardinality(), property=r.get_property(),
                                       filler=self._simplify(r.get_filler()))

    @_simplify.register
    def _(self, r: OWLObjectHasSelf) -> OWLObjectHasSelf:
        return r

    @_simplify.register
    def _(self, r: OWLObjectHasValue):
        return r

    @_simplify.register
    def _(self, r: OWLObjectOneOf):
        return OWLObjectOneOf(_sort_by_ordered_owl_object(set(r.individuals())))



""" Test DRILL"""
#import random
#import pandas as pd

# @TODO, write such tests for Drill after, experiments are completed.
"""
from ontolearn import KnowledgeBase
from ontolearn.refinement_operators import LengthBasedRefinement
from ontolearn.rl import DrillAverage
from ontolearn.utils import setup_logging

setup_logging("logging_test.conf")

PATH_FAMILY = 'KGs/Family/family-benchmark_rich_background.owl'

drill_pretrained_model_path = 'agent_pre_trained/model.pth'
family_embeddings_path = 'embeddings/dismult_family_benchmark/instance_emb.csv'
synthetic_problems_path = 'examples/synthetic_problems.json'

with open(synthetic_problems_path) as json_file:
    settings = json.load(json_file)

kb = KnowledgeBase(path=PATH_FAMILY)
rho = LengthBasedRefinement(knowledge_base=kb)

instance_emb = pd.read_csv(family_embeddings_path, index_col=0)


class TestDrill:
    def test_drill_regression(self):
        exp_f1_scores = {'Aunt': .80392, 'Brother': 1.0,
                         'Cousin': .72626, 'Granddaughter': 1.0,
                         'Uncle': .88372, 'Grandgrandfather': 0.94444}
        model = DrillAverage(knowledge_base=kb, refinement_operator=rho,
                             terminate_on_goal=True, instance_embeddings=instance_emb)

        for str_target_concept, examples in settings['problems'].items():
            p = set(examples['positive_examples'])
            n = set(examples['negative_examples'])
            print('Target concept: ', str_target_concept)
            concepts_to_ignore = set()
            # lets inject more background info
            if str_target_concept in ['Granddaughter', 'Aunt', 'Sister']:
                concepts_to_ignore.update(
                    {'http://www.benchmark.org/family#Brother', 'Father',
                     'Grandparent'})  # Use URI, or concept with length 1.

            returned_val = model.fit(pos=p, neg=n,ignore=concepts_to_ignore)
            assert returned_val == model
            hypotheses = list(model.best_hypotheses(n=5))
            assert hypotheses[0].quality >= exp_f1_scores[str_target_concept]
            assert hypotheses[0].quality >= hypotheses[1].quality
"""


""" Test for refinement_operators.py"""
from pytest import mark

import json

from ontolearn import KnowledgeBase
from ontolearn.utils import setup_logging
from owlapy.render import DLSyntaxObjectRenderer
from ontolearn.refinement_operators import CustomRefinementOperator, ModifiedCELOERefinement, LengthBasedRefinement, \
    ExpressRefinement
from ontolearn.search import Node

setup_logging("logging_test.conf")

with open('examples/synthetic_problems.json') as json_file:
    settings = json.load(json_file)
# because '../KGs/Family/family-benchmark_rich_background.owl'
kb = KnowledgeBase(path=settings['data_path'][3:])


def test_celoe_refinement_operator():
    r = DLSyntaxObjectRenderer()
    rho = ModifiedCELOERefinement(kb)
    for _ in enumerate(rho.refine(kb.thing, max_length=10, current_domain=kb.thing)):
        print(r.render(_[1]))
        pass


def test_length_refinement_operator():
    r = DLSyntaxObjectRenderer()
    rho = LengthBasedRefinement(kb)
    for _ in enumerate(rho.refine(kb.thing)):
        print(r.render(_[1]))
        pass


@mark.xfail
def test_custom_refinement_operator():
    r = DLSyntaxObjectRenderer()
    rho = CustomRefinementOperator(kb)
    for _ in enumerate(rho.refine(kb.thing)):
        print(r.render(_[1]))
        pass


def test_express_refinement_operator():
    r = DLSyntaxObjectRenderer()
    rho = ExpressRefinement(kb)
    for _ in enumerate(rho.refine(kb.thing)):
        print(r.render(_[1]))
        pass


if __name__ == '__main__':
    for fn in [
        test_celoe_refinement_operator,
        test_length_refinement_operator,
        test_express_refinement_operator,
        test_custom_refinement_operator,
    ]:
        print(fn)
        fn()



import datetime
import os
import pickle
import random
import time
from typing import Callable, Set, TypeVar

from ontolearn.utils.log_config import setup_logging  # noqa: F401
from owlapy.model import OWLNamedIndividual, IRI, OWLClass, HasIRI

Factory = Callable

# DEFAULT_FMT = '[{elapsed:0.8f}s] {name}({args}) -> {result}'
DEFAULT_FMT = 'Func:{name} took {elapsed:0.8f}s'
flag_for_performance = False


def parametrized_performance_debugger(fmt=DEFAULT_FMT):
    def decorate(func):
        if flag_for_performance:
            def clocked(*_args):
                t0 = time.time()
                _result = func(*_args)
                elapsed = time.time() - t0
                name = func.__name__
                args = ', '.join(repr(arg) for arg in _args)
                result = repr(_result)
                print(fmt.format(**locals()))
                return _result

            return clocked
        else:
            return func

    return decorate


def performance_debugger(func_name):
    def function_name_decorator(func):
        def debug(*args, **kwargs):
            start = time.time()
            r = func(*args, **kwargs)
            print(func_name, ' took ', round(time.time() - start, 4), ' seconds')

            return r

        return debug

    return function_name_decorator


def create_experiment_folder(folder_name='Log'):
    from ontolearn.utils import log_config
    if log_config.log_dirs:
        path_of_folder = log_config.log_dirs[-1]
    else:
        directory = os.getcwd() + '/' + folder_name + '/'
        folder_name = datetime.datetime.now().strftime("%Y%m%d_%H%M%S_%f")
        path_of_folder = directory + folder_name
        os.makedirs(path_of_folder)
    return path_of_folder, path_of_folder[:path_of_folder.rfind('/')]


def serializer(*, object_: object, path: str, serialized_name: str):
    with open(path + '/' + serialized_name + ".p", "wb") as f:
        pickle.dump(object_, f)
    f.close()


def deserializer(*, path: str, serialized_name: str):
    with open(path + "/" + serialized_name + ".p", "rb") as f:
        obj_ = pickle.load(f)
    f.close()
    return obj_


def apply_TSNE_on_df(df) -> None:
    from sklearn.manifold import TSNE
    from matplotlib import pyplot as plt
    low_emb = TSNE(n_components=2).fit_transform(df)
    plt.scatter(low_emb[:, 0], low_emb[:, 1])
    plt.title('Instance Representatons via TSNE')
    plt.show()


def balanced_sets(a: set, b: set) -> (set, set):
    """
    Balance given two sets through sampling without replacement.
    Returned sets have the same length.
    @param a:
    @param b:
    @return:
    """

    if len(a) > len(b):
        sampled_a = random.sample(a, len(b))
        return set(sampled_a), b
    elif len(b) > len(a):
        sampled_b = random.sample(b, len(a))
        return a, set(sampled_b)
    else:
        assert len(a) == len(b)
        return a, b


def read_csv(path):
    """
    Path leads a folder containing embeddings in csv format.
    indexes correspond subjects or predicates or objects in n-triple.
    @param path:
    @return:
    """
    import pandas as pd
    assertion_path_isfile(path)
    df = pd.read_csv(path, index_col=0)
    assert (df.all()).all()  # all columns and all rows are not none.
    return df


def assertion_path_isfile(path) -> None:
    try:
        assert path is not None
    except AssertionError:
        print(f'Path can not be:{path}')
        raise

    try:
        assert os.path.isfile(path)
    except (AssertionError, TypeError):
        print(f'Input:{path} not found.')
        raise


def sanity_checking_args(args):
    try:
        assert os.path.isfile(args.path_knowledge_base)
    except AssertionError:
        print(f'--path_knowledge_base ***{args.path_knowledge_base}*** does not lead to a file.')
        raise

    try:
        assert os.path.isfile(args.path_knowledge_base_embeddings)
    except AssertionError:
        print(f'--path_knowledge_base_embeddings ***{args.path_knowledge_base_embeddings}*** does not lead to a file.')
        raise

    assert args.min_length > 0
    assert args.max_length > 0
    assert args.min_num_concepts > 0
    assert args.min_num_concepts > 0
    assert args.min_num_instances_per_concept > 0
    assert os.path.isfile(args.path_knowledge_base)
    if hasattr(args, 'num_fold_for_k_fold_cv'):
        assert args.num_fold_for_k_fold_cv > 0
    if hasattr(args, 'max_test_time_per_concept'):
        assert args.max_test_time_per_concept > 1

    if hasattr(args, 'num_of_sequential_actions'):
        assert args.num_of_sequential_actions > 0

    if hasattr(args, 'batch_size'):
        assert args.batch_size > 1


_T = TypeVar('_T', bound=HasIRI)


def _read_iri_file(file: str, type_: Factory[[IRI], _T]) -> Set[_T]:
    """Read a text file containing IRIs (one per line) and return the content as a set of instances created by the
    given type

    Args:
        file: path to the text file with the IRIs of the named individuals
        type_: factory or type to create from the IRI

    Returns:
        set of type_ instances with these IRIs
    """

    def optional_angles(iri: str):
        if iri.startswith('<'):
            return iri[1:-1]
        else:
            return iri

    with open(file, 'r') as f:
        inds = map(type_,
                   map(IRI.create,
                       map(optional_angles,
                           f.read().splitlines())))
    return set(inds)


def read_individuals_file(file: str) -> Set[OWLNamedIndividual]:
    """Read a text file containing IRIs of Named Individuals (one per line) and return the content as a set of OWL
    Named Individuals

    Args:
        file: path to the text file with the IRIs of the named individuals

    Returns:
        set of OWLNamedIndividual with these IRIs
    """
    return _read_iri_file(file, OWLNamedIndividual)


def read_named_classes_file(file: str) -> Set[OWLClass]:
    """Read a text file containing IRIs of OWL Named Classes (one per line) and return the content as a set of OWL
    Classes

    Args:
        file: path to the text file with the IRIs of the classes

    Returns:
        set of OWLNamedIndividual with these IRIs
    """
    return _read_iri_file(file, OWLClass)



import inspect
import logging
from typing import Iterable, TypeVar

from ontolearn.abstracts import AbstractHeuristic, AbstractScorer, BaseRefinement, AbstractKnowledgeBase, \
    AbstractLearningProblem, AbstractNode
from ontolearn.base_concept_learner import BaseConceptLearner

logger = logging.getLogger(__name__)


def _get_matching_opts(_Type, optargs, kwargs, *, prefix=None):
    """find the keys in kwargs that are parameters of _Type

    if prefix is specified, the keys in kwargs need to be prefixed with prefix_
    """
    opts = {}
    if prefix is None:
        def p(s):
            return s
    else:
        def p(s):
            return prefix + "_" + s
    sig = set()
    sig.update(inspect.signature(_Type).parameters.keys())
    sig.difference_update({'args', 'kwds'})
    try:
        sig.update(inspect.signature(_Type.__init__).parameters.keys())
        sig.discard('self')
    except AttributeError:
        pass

    for opt in sig:
        if p(opt) in kwargs:
            opts[opt] = kwargs.pop(p(opt))
        elif opt in optargs:
            opts[opt] = optargs[opt]
    return opts


_N = TypeVar('_N', bound=AbstractNode)  #:


def ModelAdapter(*args, **kwargs):  # noqa: C901
    """Create a new Concept learner through the model adapter

    Args:
        knowledge_base (AbstractKnowledgeBase): a knowledge base
        knowledge_base_type: a knowledge base type
        ...: knowledge base arguments
        refinement_operator_type: a refinement operator type
        ...: refinement operator arguments
        quality_type: an Abstract Scorer type
        ...: quality arguments
        heuristic_func (AbstractHeuristic): a heuristic
        heuristic_type: an Abstract Heuristic type
        ...: arguments for the heuristic type
        learner_type: a Base Concept Learner type
        ...: arguments for the learning algorithm
    """

    if "knowledge_base" in kwargs:
        kb = kwargs.pop("knowledge_base")
        if "knowledge_base_type" in kwargs:
            raise ValueError("both knowledge_base and _type specified")
    else:
        kb_type = kwargs.pop("knowledge_base_type", None)
        if kb_type is None:
            from ontolearn import KnowledgeBase
            kb_type = KnowledgeBase
        else:
            kb_type = kb_type

        kb_args = _get_matching_opts(kb_type, {}, kwargs)
        try:
            kb = kb_type(**kb_args)
        except TypeError:
            kb = None
    if kb is not None:
        assert isinstance(kb, AbstractKnowledgeBase)

    if "ignore" in kwargs:
        from ontolearn import KnowledgeBase
        assert isinstance(kb, KnowledgeBase)
        target_kb = kb.ignore_and_copy(ignored_classes=kwargs.pop("ignore"))
    else:
        target_kb = kb

    if "refinement_operator" in kwargs:
        operator = kwargs.pop("refinement_operator")
        if "refinement_operator_type" in kwargs:
            raise ValueError("both refinement_operator and _type specified")
    else:
        op_type = kwargs.pop("refinement_operator_type", None)
        if op_type is None:
            from ontolearn.refinement_operators import ModifiedCELOERefinement
            op_type = ModifiedCELOERefinement
        assert issubclass(op_type, BaseRefinement)
        operator = op_type(**_get_matching_opts(
            op_type, {
                'knowledge_base': target_kb
            }, kwargs))
    assert isinstance(operator, BaseRefinement)

    if "quality_func" in kwargs:
        qual = kwargs.pop("quality_func")
        if "quality_type" in kwargs:
            raise ValueError("both quality_func and _type specified")
    else:
        quality_type = kwargs.pop("quality_type", None)
        if quality_type is None:
            from ontolearn.metrics import F1
            quality_type = F1
        assert issubclass(quality_type, AbstractScorer)
        qual = quality_type(**_get_matching_opts(quality_type, {}, kwargs))
    assert isinstance(qual, AbstractScorer)

    if "heuristic_func" in kwargs:
        heur = kwargs.pop("heuristic_func")
        if "heuristic_type" in kwargs:
            raise ValueError("both heuristic_func and _type specified")
    else:
        heuristic_type = kwargs.pop("heuristic_type", None)
        if heuristic_type is None:
            from ontolearn.heuristics import CELOEHeuristic
            heuristic_type = CELOEHeuristic
        assert issubclass(heuristic_type, AbstractHeuristic)
        heur = heuristic_type(**_get_matching_opts(heuristic_type, {}, kwargs))
    assert isinstance(heur, AbstractHeuristic)

    if "learner" in kwargs:
        learner = kwargs.pop("learner")
        learner_type = type(learner)
        if "learner_type" in kwargs:
            raise ValueError("both learner and _type specified")
    else:
        learner_type = kwargs.pop("learner_type", None)
        if learner_type is None:
            from ontolearn.concept_learner import CELOE
            learner_type = CELOE
        assert issubclass(learner_type, BaseConceptLearner)
        learner_args = _get_matching_opts(learner_type, {}, kwargs)
        learner = None

    other_components = dict()
    clearkeys = set()
    for k in list(kwargs):
        if k in kwargs and k.endswith("_type"):
            clearkeys.add(k)
            cls = kwargs[k]
            assert issubclass(cls, object)
            other_components[k[:-5]] = (cls, _get_matching_opts(cls, {}, kwargs))

    for k in clearkeys:
        kwargs.pop(k)

    if kwargs:
        logger.warning("Unused parameters: %s", kwargs)

    other_instances = dict()
    for k in other_components:
        cls = other_components[k][0]
        logger.debug("Instantiating %s of type %s", k, cls)

        # noinspection PyArgumentList
        inst = cls(**_get_matching_opts(cls, {
            'knowledge_base': target_kb,
            'refinement_operator': operator,
            'quality_func': qual,
            'heuristic_func': heur,
        }, other_components[k][1]))
        other_instances[k] = inst

    if learner is None:
        learner = learner_type(**_get_matching_opts(
            learner_type, {
                **other_instances,
                'knowledge_base': target_kb,
                'refinement_operator': operator,
                'quality_func': qual,
                'heuristic_func': heur,
            }, learner_args
        ))

    return learner



""" Test the base module"""
import os

from ontolearn.knowledge_base import KnowledgeBase
from ontolearn.utils import setup_logging

setup_logging("logging_test.conf")

PATH_FAMILY = 'KGs/Family/family-benchmark_rich_background.owl'
PATH_FATHER = 'KGs/father.owl'


def test_knowledge_base():
    KnowledgeBase(path=PATH_FAMILY)
    # assert kb.name == 'family-benchmark_rich_background'
    #
    # assert kb.property_hierarchy
    # assert kb.property_hierarchy.all_properties
    # assert len(kb.property_hierarchy.all_properties) >= \
    #        len(kb.property_hierarchy.data_properties)
    # assert len(kb.property_hierarchy.all_properties) >= \
    #        len(kb.property_hierarchy.object_properties)


def test_multiple_knowledge_bases():
    KnowledgeBase(path=PATH_FAMILY)

    # There should not be an exception here
    # (that refers to the family ontology)


# def test_knowledge_base_save():
#     kb = KnowledgeBase(path=PATH_FAMILY)
#     kb.save('test_kb_save', rdf_format='nt')
#     assert os.stat('test_kb_save.nt').st_size > 0
#     os.remove('test_kb_save.nt')


if __name__ == '__main__':
    test_knowledge_base()
    test_multiple_knowledge_bases()
    # test_knowledge_base_save()



import weakref
from abc import ABCMeta, abstractmethod
from typing import Final, Union, overload
from weakref import WeakKeyDictionary

from owlapy import namespaces
from owlapy.model._base import OWLAnnotationSubject, OWLAnnotationValue
from owlapy.namespaces import Namespaces


class HasIRI(metaclass=ABCMeta):
    __slots__ = ()

    @abstractmethod
    def get_iri(self) -> 'IRI':
        """Gets the IRI of this object.

        Returns:
            The IRI of this object
        """
        pass


class _WeakCached(type):
    __slots__ = ()

    def __init__(cls, what, bases, dct):
        super().__init__(what, bases, dct)
        cls._cache = WeakKeyDictionary()

    def __call__(cls, *args, **kwargs):
        _temp = super().__call__(*args, **kwargs)
        ret = cls._cache.get(_temp)
        if ret is None:
            cls._cache[_temp] = weakref.ref(_temp)
            return _temp
        else:
            return ret()


class _meta_IRI(ABCMeta, _WeakCached):
    __slots__ = ()
    pass


class IRI(OWLAnnotationSubject, OWLAnnotationValue, metaclass=_meta_IRI):
    """An IRI, consisting of a namespace and a remainder"""
    __slots__ = '_namespace', '_remainder', '__weakref__'
    type_index: Final = 0

    _namespace: str
    _remainder: str

    def __init__(self, namespace: Union[str, Namespaces], remainder: str):
        if isinstance(namespace, Namespaces):
            namespace = namespace.ns
        else:
            assert namespace[-1] in ("/", ":", "#")
        import sys
        self._namespace = sys.intern(namespace)
        self._remainder = remainder

    @overload
    @staticmethod
    def create(namespace: Namespaces, remainder: str) -> 'IRI': ...

    @overload
    @staticmethod
    def create(namespace: str, remainder: str) -> 'IRI':
        """Creates an IRI by concatenating two strings. The full IRI is an IRI that contains the characters in
        namespace + remainder.

        Args:
            namespace: The first string
            remainder: The second string

        Returns:
            An IRI whose characters consist of prefix + suffix.
        """
        ...

    @overload
    @staticmethod
    def create(string: str) -> 'IRI':
        """Creates an IRI from the specified String.

        Args:
            string: The String that specifies the IRI

        Returns:
            The IRI that has the specified string representation.
        """
        ...

    @staticmethod
    def create(string, remainder=None) -> 'IRI':
        if remainder is not None:
            return IRI(string, remainder)
        index = 1 + max(string.rfind("/"), string.rfind(":"), string.rfind("#"))
        return IRI(string[0:index], string[index:])

    def __repr__(self):
        return f"IRI({repr(self._namespace)},{repr(self._remainder)})"

    def __eq__(self, other):
        if type(other) is type(self):
            return self._namespace is other._namespace and self._remainder == other._remainder
        return NotImplemented

    def __hash__(self):
        return hash((self._namespace, self._remainder))

    def is_nothing(self):
        """Determines if this IRI is equal to the IRI that owl:Nothing is named with.

        Returns:
            :True if this IRI is equal to <http://www.w3.org/2002/07/owl#Nothing> and otherwise False
        """
        from owlapy.vocab import OWLRDFVocabulary
        return self == OWLRDFVocabulary.OWL_NOTHING.get_iri()

    def is_thing(self):
        """Determines if this IRI is equal to the IRI that owl:Thing is named with.

        Returns:
            :True if this IRI is equal to <http://www.w3.org/2002/07/owl#Thing> and otherwise False
        """
        from owlapy.vocab import OWLRDFVocabulary
        return self == OWLRDFVocabulary.OWL_THING.get_iri()

    def is_reserved_vocabulary(self) -> bool:
        """Determines if this IRI is in the reserved vocabulary. An IRI is in the reserved vocabulary if it starts with
        <http://www.w3.org/1999/02/22-rdf-syntax-ns#> or <http://www.w3.org/2000/01/rdf-schema#> or
        <http://www.w3.org/2001/XMLSchema#> or <http://www.w3.org/2002/07/owl#>

        Returns:
            True if the IRI is in the reserved vocabulary, otherwise False.
        """
        return self._namespace == namespaces.OWL or self._namespace == namespaces.RDF \
               or self._namespace == namespaces.RDFS or self._namespace == namespaces.XSD

    def as_iri(self) -> 'IRI':
        # documented in parent
        return self

    def as_str(self) -> str:
        """
        Returns:
            the string that specifies the IRI
        """
        return self._namespace + self._remainder

    def get_short_form(self) -> str:
        """Gets the short form.

        Returns:
            A string that represents the short form.
        """
        return self._remainder

    def get_namespace(self) -> str:
        """
        Returns:
            the namespace as string
        """
        return self._namespace

    def get_remainder(self) -> str:
        """
        Returns:
            the remainder (coincident with NCName usually) for this IRI.
        """
        return self._remainder



import unittest

from owlapy.model import OWLDataMinCardinality, OWLObjectIntersectionOf, OWLObjectSomeValuesFrom, \
    OWLThing, OWLObjectComplementOf, OWLObjectUnionOf, OWLNamedIndividual, OWLObjectOneOf, OWLObjectHasValue, \
    OWLObjectMinCardinality, IRI, OWLDataProperty, DoubleOWLDatatype, OWLClass, OWLDataComplementOf, \
    OWLDataIntersectionOf, IntegerOWLDatatype, OWLDataExactCardinality, OWLDataHasValue, OWLDataAllValuesFrom, \
    OWLDataOneOf, OWLDataSomeValuesFrom, OWLDataUnionOf, OWLLiteral, OWLObjectProperty, BooleanOWLDatatype, \
    OWLDataMaxCardinality
from owlapy.model.providers import OWLDatatypeMinMaxInclusiveRestriction
from owlapy.render import DLSyntaxObjectRenderer, ManchesterOWLSyntaxOWLObjectRenderer


class Owlapy_DLRenderer_Test(unittest.TestCase):
    def test_ce_render(self):
        renderer = DLSyntaxObjectRenderer()
        NS = "http://example.com/father#"

        male = OWLClass(IRI.create(NS, 'male'))
        female = OWLClass(IRI.create(NS, 'female'))
        has_child = OWLObjectProperty(IRI(NS, 'hasChild'))
        has_age = OWLDataProperty(IRI(NS, 'hasAge'))

        c = OWLObjectUnionOf((male, OWLObjectSomeValuesFrom(property=has_child, filler=female)))
        r = renderer.render(c)
        print(r)
        self.assertEqual(r, "male ⊔ (∃ hasChild.female)")
        c = OWLObjectComplementOf(OWLObjectIntersectionOf((female,
                                                           OWLObjectSomeValuesFrom(property=has_child,
                                                                                   filler=OWLThing))))
        r = renderer.render(c)
        print(r)
        self.assertEqual(r, "¬(female ⊓ (∃ hasChild.⊤))")
        c = OWLObjectSomeValuesFrom(property=has_child,
                                    filler=OWLObjectSomeValuesFrom(property=has_child,
                                                                   filler=OWLObjectSomeValuesFrom(property=has_child,
                                                                                                  filler=OWLThing)))
        r = renderer.render(c)
        print(r)
        self.assertEqual(r, "∃ hasChild.(∃ hasChild.(∃ hasChild.⊤))")

        i1 = OWLNamedIndividual(IRI.create(NS, 'heinz'))
        i2 = OWLNamedIndividual(IRI.create(NS, 'marie'))
        oneof = OWLObjectOneOf((i1, i2))
        r = renderer.render(oneof)
        print(r)
        self.assertEqual(r, "{heinz ⊔ marie}")

        hasvalue = OWLObjectHasValue(property=has_child, individual=i1)
        r = renderer.render(hasvalue)
        print(r)
        self.assertEqual(r, "∃ hasChild.{heinz}")

        mincard = OWLObjectMinCardinality(cardinality=2, property=has_child, filler=OWLThing)
        r = renderer.render(mincard)
        print(r)
        self.assertEqual(r, "≥ 2 hasChild.⊤")

        d = OWLDataSomeValuesFrom(property=has_age,
                                  filler=OWLDataComplementOf(DoubleOWLDatatype))
        r = renderer.render(d)
        print(r)
        self.assertEqual(r, "∃ hasAge.¬xsd:double")

        datatype_restriction = OWLDatatypeMinMaxInclusiveRestriction(40, 80)

        dr = OWLDataAllValuesFrom(property=has_age, filler=OWLDataUnionOf([datatype_restriction, IntegerOWLDatatype]))
        r = renderer.render(dr)
        print(r)
        self.assertEqual(r, "∀ hasAge.(xsd:integer[≥ 40 , ≤ 80] ⊔ xsd:integer)")

        dr = OWLDataSomeValuesFrom(property=has_age,
                                   filler=OWLDataIntersectionOf([OWLDataOneOf([OWLLiteral(32.5), OWLLiteral(4.5)]),
                                                                 IntegerOWLDatatype]))
        r = renderer.render(dr)
        print(r)
        self.assertEqual(r, "∃ hasAge.({32.5 ⊔ 4.5} ⊓ xsd:integer)")

        hasvalue = OWLDataHasValue(property=has_age, value=OWLLiteral(50))
        r = renderer.render(hasvalue)
        print(r)
        self.assertEqual(r, "∃ hasAge.{50}")

        exactcard = OWLDataExactCardinality(cardinality=1, property=has_age, filler=IntegerOWLDatatype)
        r = renderer.render(exactcard)
        print(r)
        self.assertEqual(r, "= 1 hasAge.xsd:integer")

        maxcard = OWLDataMaxCardinality(cardinality=4, property=has_age, filler=DoubleOWLDatatype)
        r = renderer.render(maxcard)
        print(r)
        self.assertEqual(r, "≤ 4 hasAge.xsd:double")

        mincard = OWLDataMinCardinality(cardinality=7, property=has_age, filler=BooleanOWLDatatype)
        r = renderer.render(mincard)
        print(r)
        self.assertEqual(r, "≥ 7 hasAge.xsd:boolean")


class Owlapy_ManchesterRenderer_Test(unittest.TestCase):
    def test_ce_render(self):
        renderer = ManchesterOWLSyntaxOWLObjectRenderer()
        NS = "http://example.com/father#"

        male = OWLClass(IRI.create(NS, 'male'))
        female = OWLClass(IRI.create(NS, 'female'))
        has_child = OWLObjectProperty(IRI(NS, 'hasChild'))
        has_age = OWLDataProperty(IRI(NS, 'hasAge'))

        c = OWLObjectUnionOf((male, OWLObjectSomeValuesFrom(property=has_child, filler=female)))
        r = renderer.render(c)
        print(r)
        self.assertEqual(r, "male or (hasChild some female)")
        c = OWLObjectComplementOf(OWLObjectIntersectionOf((female,
                                                           OWLObjectSomeValuesFrom(property=has_child,
                                                                                   filler=OWLThing))))
        r = renderer.render(c)
        print(r)
        self.assertEqual(r, "not (female and (hasChild some Thing))")
        c = OWLObjectSomeValuesFrom(property=has_child,
                                    filler=OWLObjectSomeValuesFrom(property=has_child,
                                                                   filler=OWLObjectSomeValuesFrom(property=has_child,
                                                                                                  filler=OWLThing)))
        r = renderer.render(c)
        print(r)
        self.assertEqual(r, "hasChild some (hasChild some (hasChild some Thing))")

        i1 = OWLNamedIndividual(IRI.create(NS, 'heinz'))
        i2 = OWLNamedIndividual(IRI.create(NS, 'marie'))
        oneof = OWLObjectOneOf((i1, i2))
        r = renderer.render(oneof)
        print(r)
        self.assertEqual(r, "{heinz , marie}")

        hasvalue = OWLObjectHasValue(property=has_child, individual=i1)
        r = renderer.render(hasvalue)
        print(r)
        self.assertEqual(r, "hasChild value heinz")

        mincard = OWLObjectMinCardinality(cardinality=2, property=has_child, filler=OWLThing)
        r = renderer.render(mincard)
        print(r)
        self.assertEqual(r, "hasChild min 2 Thing")

        d = OWLDataSomeValuesFrom(property=has_age,
                                  filler=OWLDataComplementOf(DoubleOWLDatatype))
        r = renderer.render(d)
        print(r)
        self.assertEqual(r, "hasAge some not xsd:double")

        datatype_restriction = OWLDatatypeMinMaxInclusiveRestriction(40, 80)

        dr = OWLDataAllValuesFrom(property=has_age, filler=OWLDataUnionOf([datatype_restriction, IntegerOWLDatatype]))
        r = renderer.render(dr)
        print(r)
        self.assertEqual(r, "hasAge only (xsd:integer[≥ 40 , ≤ 80] or xsd:integer)")

        dr = OWLDataSomeValuesFrom(property=has_age,
                                   filler=OWLDataIntersectionOf([OWLDataOneOf([OWLLiteral(32.5), OWLLiteral(4.5)]),
                                                                 IntegerOWLDatatype]))
        r = renderer.render(dr)
        print(r)
        self.assertEqual(r, "hasAge some ({32.5 , 4.5} and xsd:integer)")

        hasvalue = OWLDataHasValue(property=has_age, value=OWLLiteral(50))
        r = renderer.render(hasvalue)
        print(r)
        self.assertEqual(r, "hasAge value 50")

        maxcard = OWLDataExactCardinality(cardinality=1, property=has_age, filler=IntegerOWLDatatype)
        r = renderer.render(maxcard)
        print(r)
        self.assertEqual(r, "hasAge exactly 1 xsd:integer")

        maxcard = OWLDataMaxCardinality(cardinality=4, property=has_age, filler=DoubleOWLDatatype)
        r = renderer.render(maxcard)
        print(r)
        self.assertEqual(r, "hasAge max 4 xsd:double")

        mincard = OWLDataMinCardinality(cardinality=7, property=has_age, filler=BooleanOWLDatatype)
        r = renderer.render(mincard)
        print(r)
        self.assertEqual(r, "hasAge min 7 xsd:boolean")


if __name__ == '__main__':
    unittest.main()



import unittest
from typing import TypeVar

from ontolearn.core.owl.hierarchy import ClassHierarchy, ObjectPropertyHierarchy, AbstractHierarchy
from ontolearn.utils import setup_logging
from owlapy.model import OWLClass, OWLObjectProperty, IRI
from owlapy.owlready2 import OWLOntologyManager_Owlready2, OWLReasoner_Owlready2

_T = TypeVar('_T')  #:

setup_logging("logging_test.conf")


class Owl_Core_PropertyHierarchy_Test(unittest.TestCase):
    def test_object_property_hierarchy(self):
        NS = "http://www.biopax.org/examples/glycolysis#"
        mgr = OWLOntologyManager_Owlready2()
        onto = mgr.load_ontology(IRI.create("file://KGs/Biopax/biopax.owl"))
        reasoner = OWLReasoner_Owlready2(onto)

        oph = ObjectPropertyHierarchy(reasoner)
        # for k in sorted(oph.roots()):
        #     _print_children(oph, k)

        participants = OWLObjectProperty(IRI.create(NS, 'PARTICIPANTS'))
        target_props = frozenset({OWLObjectProperty(IRI(NS, 'COFACTOR')),
                                  OWLObjectProperty(IRI(NS, 'CONTROLLED')),
                                  OWLObjectProperty(IRI(NS, 'CONTROLLER')),
                                  OWLObjectProperty(IRI(NS, 'LEFT')),
                                  OWLObjectProperty(IRI(NS, 'RIGHT'))})
        self.assertEqual(frozenset(oph.sub_object_properties(participants)), target_props)


class Owl_Core_ClassHierarchy_Test(unittest.TestCase):
    # TODO
    # def test_class_hierarchy_circle(self):
    #     NS = "http://example.org/circle#"
    #     a1 = OWLClass(IRI.create(NS, 'A1'))
    #     a3 = OWLClass(IRI.create(NS, 'A3'))
    #     a2 = OWLClass(IRI.create(NS, 'A2'))
    #     a0 = OWLClass(IRI.create(NS, 'A0'))
    #     a4 = OWLClass(IRI.create(NS, 'A4'))
    #     ch = ClassHierarchy({a1: [a2],
    #                          a2: [a3],
    #                          a3: [a1, a4],
    #                          a0: [a1],
    #                          a4: []}.items())
    #     for k in sorted(ch.roots()):
    #         _print_children(ch, k)

    def test_class_hierarchy_restrict(self):
        NS = "http://www.benchmark.org/family#"
        mgr = OWLOntologyManager_Owlready2()
        onto = mgr.load_ontology(IRI.create("file://KGs/Family/family-benchmark_rich_background.owl"))
        reasoner = OWLReasoner_Owlready2(onto)

        ch = ClassHierarchy(reasoner).restrict_and_copy(remove=frozenset({OWLClass(IRI(NS, 'Grandchild'))}))

        target_cls = frozenset({OWLClass(IRI(NS, 'Daughter')),
                                OWLClass(IRI(NS, 'Granddaughter')),
                                OWLClass(IRI(NS, 'Grandson')),
                                OWLClass(IRI(NS, 'Son'))})
        self.assertEqual(frozenset(ch.sub_classes(OWLClass(IRI(NS, 'Child')))), target_cls)

    def test_class_hierarchy_children(self):
        NS = "http://example.com/father#"
        mgr = OWLOntologyManager_Owlready2()
        onto = mgr.load_ontology(IRI.create("file://KGs/father.owl"))
        reasoner = OWLReasoner_Owlready2(onto)

        ch = ClassHierarchy(reasoner)
        # for k in sorted(ch.roots()):
        #     _print_children(ch, k)

        target_cls = frozenset({OWLClass(IRI(NS, 'female')),
                                OWLClass(IRI(NS, 'male'))})
        self.assertEqual(frozenset(ch.sub_classes(OWLClass(IRI(NS, 'person')))), target_cls)

    def test_class_hierarchy_parents_roots(self):
        NS = "http://www.benchmark.org/family#"
        mgr = OWLOntologyManager_Owlready2()
        onto = mgr.load_ontology(IRI.create("file://KGs/Family/family-benchmark_rich_background.owl"))
        reasoner = OWLReasoner_Owlready2(onto)

        ch = ClassHierarchy(reasoner)
        grandmother = OWLClass(IRI(NS, 'Grandmother'))
        # _print_parents(ch, grandmother)

        target_cls = frozenset({OWLClass(IRI(NS, 'Female')),
                                OWLClass(IRI(NS, 'Grandparent'))})
        self.assertEqual(frozenset(ch.super_classes(grandmother)), target_cls)

        target_cls = frozenset({OWLClass(IRI(NS, 'Person'))})
        self.assertEqual(frozenset(ch.roots()), target_cls)

    def test_class_hierarchy_siblings(self):
        NS = "http://www.benchmark.org/family#"
        mgr = OWLOntologyManager_Owlready2()
        onto = mgr.load_ontology(IRI.create("file://KGs/Family/family-benchmark_rich_background.owl"))
        reasoner = OWLReasoner_Owlready2(onto)

        ch = ClassHierarchy(reasoner)
        child = OWLClass(IRI(NS, 'Child'))
        target_cls = frozenset({OWLClass(IRI(NS, 'Parent')),
                                OWLClass(IRI(NS, 'PersonWithASibling')),
                                OWLClass(IRI(NS, 'Female')),
                                OWLClass(IRI(NS, 'Male'))})
        self.assertEqual(frozenset(ch.siblings(child)), target_cls)

    def test_class_hierarchy_leaves(self):
        NS = "http://www.benchmark.org/family#"
        mgr = OWLOntologyManager_Owlready2()
        onto = mgr.load_ontology(IRI.create("file://KGs/Family/family-benchmark_rich_background.owl"))
        reasoner = OWLReasoner_Owlready2(onto)

        ch = ClassHierarchy(reasoner)
        # for k in sorted(ch.roots()):
        #     _print_children(ch, k)
        children = OWLClass(IRI(NS, 'Child'))
        target_leaves = frozenset({OWLClass(IRI(NS, 'Daughter')),
                                   OWLClass(IRI(NS, 'Granddaughter')),
                                   OWLClass(IRI(NS, 'Grandson')),
                                   OWLClass(IRI(NS, 'Son'))})
        leaves = frozenset(ch.leaves(children))
        self.assertEqual(leaves, target_leaves)


# debug functions

def _print_children(hier: AbstractHierarchy[_T], c: _T, level: int = 0) -> None:
    print(' ' * 2 * level, c, '=>')
    for d in sorted(hier.children(c)):
        _print_children(hier, d, level + 1)


def _print_parents(hier: AbstractHierarchy[_T], c: _T, level: int = 0) -> None:
    print(' ' * 2 * level, c, '<=')
    for d in sorted(hier.parents(c)):
        _print_parents(hier, d, level + 1)


if __name__ == '__main__':
    unittest.main()
